from math import pi
from nxt.motcont import MotCont
from nxt.bluesock import BlueSock
from nxt.locator import find_one_brick
from nxt.sensor import PORT_1, PORT_2, PORT_3, PORT_4, Touch, Light, Ultrasonic
from nxt.sensor.generic import Color20
from nxt.sensor.hitechnic import Compass, Accelerometer
from time import sleep


"""
GLOBALS
"""
BRICK = None
MOTOR_CONTROL = None
COLORSENSOR = None

    
def connect(brick_id_address = None):
    global BRICK, MOTOR_CONTROL
    if brick_id_address:
        BRICK = BlueSock(brick_id_address).connect()
    else:
        BRICK = find_one_brick(debug=True)
    MOTOR_CONTROL = MotCont(BRICK)
    MOTOR_CONTROL.start()

def close():
    MOTOR_CONTROL.stop()
    BRICK.sock.close()
    
def get_info():
    return BRICK.get_device_info()

def play_file(sound, times):
    for i in range(0, times):
        BRICK.play_sound_file(False, sound + '.rso')
        sleep(1)

def play_tone(frequency, duration):
    BRICK.play_tone(frequency, duration)

def get_touch1_sample(port = PORT_1):
    return Touch(BRICK, port).get_sample()
     
def get_touch2_sample(port = PORT_2):
    return Touch(BRICK, port).get_sample()

def activate_color_sensor(port = PORT_3 ):
    global COLORSENSOR
    COLORSENSOR = Color20(BRICK, port)

def get_color_sample():
    """
    Only works if really close to ground (1cm) otherwise it can fail and mix colors up.
    Black = 1
    Blue = 2
    Green = 3
    Yellow = 4
    Red = 5
    White = 6
    
    If using light color mode red green blue
    Dark brightness is as low as it goes 0 - 70 pretty dark
    Light brightness is as high as it goes around 650
    """
    return COLORSENSOR.get_reflected_light(COLORSENSOR.get_light_color())

def set_color(color):
    """
    all = Color sensor mode 
          (Only sees pure basic colors look at get_color_sample() 
          comment for readings).
    red, green and blue = Light color sensor mode 
                          Measures brightnes and darkness but you have to calibrate it yourself.
                          Read a value from floor then read value from dark line then ta genomsnitts av dem
                          detta ar cut-off.
    """
    if color.lower() == "red":
        _color = 0x0E
    elif color.lower() == "green":
        _color = 0x0F 
    elif color.lower() == "blue":
        _color = 0x10
    elif color.lower() == "all":
        _color = 0x0D
    elif color.lower() == "none":
        _color = 0x11
    else:
        print("Only supports 'red', 'green', 'blue', 'all', 'none' in colors")
    
    COLORSENSOR.set_light_color(_color)

def get_color():
    color = COLORSENSOR.get_light_color()
    if  color == 0x0E:
        _color = "red"
    elif color == 0x0F:
        _color = "green"
    elif color == 0x10:
        _color = "blue"
    elif color == 0x0D:
        _color = "all"
    elif color == 0x11:
        _color = "none"
        
    return _color
  
def get_ultrasonic_sample(port = PORT_4):
    return Ultrasonic(BRICK, port).get_sample()

def get_compass_sample(port = PORT_2):
    """Returns heading from North in degrees. 0 to 360"""
    return Compass(BRICK, port).get_sample()

def get_relative_compass(target = 0, port = PORT_2):
    """Returns how far away from target you are if you have 360 as target and you are on 340 you are 20 away from target, -180 to 180"""
    return Compass(BRICK, port).get_relative_heading(target)

def is_in_range_compass(minValue, maxValue, port = PORT_2):
    """Returns True if compass in between the values given"""
    return Compass(BRICK, port).is_in_range(minValue, maxValue)

def get_x_acceleration(port = PORT_1):
    return Accelerometer(BRICK, port).get_sample().x

def get_y_acceleration(port = PORT_1):
    return Accelerometer(BRICK, port).get_sample().y 

def get_z_acceleration(port = PORT_1):
    return Accelerometer(BRICK, port).get_sample().z


def run_motor_A(power, tacho_units):
    MOTOR_CONTROL.cmd(0, power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    is_ready(0)
    
def run_motor_B(power, tacho_units):
    MOTOR_CONTROL.cmd(1, power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    is_ready(1)
    
def run_motor_C(power, tacho_units):
    MOTOR_CONTROL.cmd(2, power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    is_ready(2)
    
def move(power, tacho_units):
    MOTOR_CONTROL.cmd(5, power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    is_ready(5)

def move_forever(power):
    MOTOR_CONTROL.set_output_state(5, power, 0, speedreg=0)

def stop_move_forever():
    MOTOR_CONTROL.set_output_state(5, 0, 0, speedreg=0)
    is_ready(5)

def turn_left_forever(power):
    MOTOR_CONTROL.set_output_state(1, power, 0, speedreg=0)

def turn_right_forever(power):
    MOTOR_CONTROL.set_output_state(2, power, 0, speedreg=0)
        
def turn_right(power, tacho_units):   
    MOTOR_CONTROL.cmd(2, power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    MOTOR_CONTROL.cmd(1, -power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    is_ready(5)
        
def turn_left(power, tacho_units):
    MOTOR_CONTROL.cmd(1, power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    MOTOR_CONTROL.cmd(2, -power, tacho_units, speedreg=0, smoothstart=1, brake=0)
    is_ready(5)
        
def is_ready(port):
    if port < 3:
        while(not MOTOR_CONTROL.is_ready(port)):
            pass
    else:
        if port == 3:
            port = (0,1)
        elif port == 4:
            port = (0,2)
        else: 
            port = (1,2)
            
        while(True):
            if MOTOR_CONTROL.is_ready(port[0]) and MOTOR_CONTROL.is_ready(port[1]):
                break